from typing import TypeVar, Generic, Callable, Iterable
import os
from time import time
from fastapi import UploadFile, Request, HTTPException, Depends
from dslog import Logger
import pure_cv as vc
from chess_pairings import GameId

def token(req: Request) -> str:
  auth = req.headers.get('Authorization') or ''
  if (parts := auth.split(' ')) and len(parts) == 2:
    return parts[1]
  raise HTTPException(status_code=401)

Bearer: str = Depends(token)

async def notify(gameId: GameId, logger: Logger):
  endpoint = os.getenv('NOTIFY_ENDPOINT')
  token = os.getenv('NOTIFY_TOKEN')
  try:
    if endpoint is None:
      raise ValueError('No NOTIFY_ENDPOINT environment variable')
    import aiohttp
    async with aiohttp.ClientSession() as session:
      async with session.post(endpoint, json=gameId, headers=[('Authorization', f'Bearer {token}')]) as r: # type: ignore
        if r.status == 200:
          logger(f'Notified "{endpoint}" of game {gameId}', level='DEBUG')
        else:
          logger(f'Failed to notify "{endpoint}" of game {gameId}: {r.status}. Content:', await r.text(), level='ERROR')
  except Exception as e:
    logger(f'Exception notifying "{endpoint}" of game {gameId}:', e, level='ERROR')


def safe_extension(file: UploadFile) -> str | None:
  if file.filename and len(parts := file.filename.split('.')) == 2:
    return parts[1]
  
def descale2jpg(img: bytes, height: int):
  mat = vc.decode(img)
  return vc.encode(vc.descale_h(mat, height), format='.jpg')

K = TypeVar('K')
T = TypeVar('T')

class TimedIterableCache(Generic[K, T]):
  def __init__(self, get: Callable[[K], Iterable[T]], *, ttl_secs: float, max_entries: int):
    self._get = get
    self.ttl_secs = ttl_secs
    self._cache: dict[K, tuple[list[T], float]] = {}
    self.max_entries = max_entries

  def insert(self, k: K, access_time: float | None = None) -> Iterable[T]:
    access_time = access_time or time()
    if len(self._cache) >= self.max_entries:
      older_client = min(self._cache.keys(), key=self._cache.__getitem__)
      del self._cache[older_client]
    
    def gen():
      stored = []
      for x in self._get(k):
        stored.append(x)
        yield x
      self._cache[k] = stored, access_time

    yield from gen()


  def __getitem__(self, key: K) -> Iterable[T]:
    now = time()
    if key not in self._cache or now - self._cache[key][1] > self.ttl_secs:
      yield from self.insert(key, now)
    else:
      yield from self._cache[key][0]


from typing import Sequence
from datetime import datetime, timedelta
from sqlmodel import select, Session
from haskellian import iter as I, Left, Right
import chess_pairings as cp
from moveread.tnmt import Round, Pairings, Game, Tournament, queries

def current_pairings(session: Session, now: datetime | None = None) -> Sequence[tuple[Tournament, Pairings]]:
  """Ongoing tournaments with available pairings"""
  now = now or datetime.now()
  tomorrow = now + timedelta(days=1)
  yesterday = now - timedelta(days=1)
  stmt = select(Tournament, Pairings) \
    .where(Tournament.start_date < tomorrow, yesterday < Tournament.end_date) \
    .join(Pairings, Tournament.tournId == Pairings.tournId) # type: ignore
  return session.exec(stmt).all()

def round_paired(session: Session, tournId: str, group: str, round: str) -> bool:
  """Does the round have pairings?"""
  stmt = queries.select_round_games(tournId, group, round)
  return session.exec(stmt).first() is not None

def round_finished(session: Session, tournId: str, group: str, round: str) -> bool:
  """Does the round have all results? (WARNING: will return true if unpaired! - by vacuity)"""
  stmt = queries.select_round_games(tournId, group, round).where(Game.result == None)
  return session.exec(stmt).first() is None

def select_paired_rounds(tournId: str, group: str):
  """Select all rounds with pairings"""
  subquery = (
    select(Game.round, Round.start_dtime)
    .where(Game.tournId == tournId, Game.group == group)
    .join(Round, (Game.tournId == Round.tournId) & (Game.group == Round.group) & (Game.round == Round.name)) # type: ignore
    .order_by(Round.start_dtime) # type: ignore
    .distinct()
    .subquery()
  )
  s: str = subquery.c.round # type: ignore
  return select(s)

def current_round(
  session: Session, tournId: str, group: str, *, now: datetime | None = None,
  round_paired = round_paired, round_finished = round_finished
):
  """Determine the current round (whose pairings should be polled)"""
  now = now or datetime.now()
  if not (rounds := session.exec(queries.select_rounds(tournId, group)).all()):
    return Left(f'No rounds found for group "{tournId}/{group}"')
  
  if (prev := I.find_last_idx(lambda r: r.start_dtime < now, rounds)) is None:
    return Right(rounds[0].name)
  
  # if next round (which hasn't yet started) is paired, nothing to do
  if prev+1 < len(rounds) and round_paired(session, **cp.roundId(tournId, group, rounds[prev+1].name)):
    return Right(None)
  
  # if current round is finished, poll the next round if it exists
  if round_finished(session, **cp.roundId(tournId, group, rounds[prev].name)):
    if prev+1 < len(rounds):
      return Right(rounds[prev+1].name)
    else:
      return Right(None)
  
  # otherwise, poll the current round
  return Right(rounds[prev].name)


from dataclasses import dataclass
from sqlmodel import Session
from moveread.tnmt import Tournament, Round, SheetModel, queries, Upload

@dataclass
class Select:
  session: Session

  def tnmt(self, tournId: str):
    return self.session.get(Tournament, tournId)
  
  def round(self, tournId: str, group: str, round: str):
    return self.session.get(Round, (tournId, group, round))
  
  def game(self, tournId: str, group: str, round: str, board: str):
    return self.session.exec(queries.select_game(tournId, group, round, board)).first()
  
  def game_uploads(self):
    return self.session.exec(queries.game_uploads()).all()
  
  def model(self, tournId: str):
    if (obj := self.session.get(SheetModel, tournId)):
      return obj.model

@dataclass
class API:
  session: Session

  @property
  def select(self):
    return Select(self.session)


from typing_extensions import Iterable, Unpack, Sequence, Coroutine, Any
from dataclasses import dataclass
import asyncio
from uuid import uuid4
from sqlalchemy import Engine
from sqlmodel import Session, select
from fastapi import UploadFile
from haskellian import either as E, Left, Either, promise as P, Iter
from kv import KV, InexistentItem
import chess_pairings as cp
from moveread.tnmt import Token, Tournament, Group, FrontendGame, queries, Game, FrontendPGN, \
  Image, PGN, Upload
from .util import TimedIterableCache, safe_extension, descale2jpg
from .pgns import export_all


@dataclass
class SDK:
  images: KV[bytes]
  engine: Engine
  cache_ttl: float = 30
  cache_max_entries: int = 1000

  def __post_init__(self):
    self.pgns_cache = TimedIterableCache(self.gen_round_pgn, ttl_secs=self.cache_ttl, max_entries=self.cache_max_entries)

  def authorize(self, token: str, tournId: str) -> bool:
    with Session(self.engine) as ses:
      results = ses.exec(select(Token).where(Token.token == token, Token.tournId == tournId))
      return results.first() is not None
    
  def get_tnmts(self, skip: int = 0, take: int = 10) -> Iterable[Tournament]:
    with Session(self.engine) as ses:
      ord: Any = Tournament.start_date.desc() # type: ignore (yeah)
      stmt = select(Tournament).order_by(ord) \
        .offset(skip).limit(take)
      return ses.exec(stmt).all()

  def get_tnmt(self, tournId: str) -> Tournament | None:
    with Session(self.engine) as ses:
      return ses.get(Tournament, tournId)
    
  def get_group(self, tournId: str, group: str) -> Group | None:
    with Session(self.engine) as ses:
      rounds = ses.exec(queries.select_paired_rounds(tournId, group)).all()
      return Group(tournId=tournId, name=group, rounds=rounds)
  
  def get_round(self, **roundId: Unpack[cp.RoundId]) -> Iterable[FrontendGame]:
    with Session(self.engine) as ses:
      stmt = select(Game).where(*queries.round_games(**roundId)).order_by(Game.index) # type: ignore
      for g in ses.exec(stmt):
        yield FrontendGame.of(g)

  def gen_round_pgn(self, roundId: tuple[str, str, str]):
    rid = cp.roundId(*roundId)
    with Session(self.engine) as ses:
      round = ses.exec(queries.select_round(**rid)).first()
      stmt = select(Game).where(*queries.round_games(**rid)).order_by(Game.index) # type: ignore
      games = ses.exec(stmt).all()
      tnmt = ses.exec(queries.select_tnmt(rid['tournId'])).first()
      yield from export_all(games, tnmt, round)
    
  def get_round_pgn(self, *, skip: int = 0, take: int | None = None, tournId: str, group: str, round: str, no_cache: bool = False) -> Iterable[str]:
    id = (tournId, group, round)
    iter = self.pgns_cache[id] if not no_cache else self.pgns_cache.insert(id)
    iter = Iter(iter).skip(skip)
    return iter.take(take) if take is not None else iter
  
  def get_group_pgn(self, *, skip: int = 0, take: int | None = None, no_cache: bool = False, **groupId: Unpack[cp.GroupId]) -> Iterable[str]:
    with Session(self.engine) as ses:
      rounds = ses.exec(queries.select_paired_rounds(**groupId)).all()
      for round in rounds:
        yield from self.get_round_pgn(skip=skip, take=take, **groupId, round=round, no_cache=no_cache)
  
  def get_pgn(self, **gameId: Unpack[cp.GameId]) -> FrontendPGN | None:
    with Session(self.engine) as ses:
      game = ses.exec(queries.select_game(**gameId)).first()
      if game and game.upload and game.upload.pgn:
        return game.upload.pgn
      
  def post_pgn(self, pgn: Sequence[str], **gameId: Unpack[cp.GameId]):
    with Session(self.engine) as ses:
      if (game := ses.exec(queries.select_game(**gameId)).first()):
        game.upload = Upload(
          id=f'{cp.stringifyId(**gameId)}_{uuid4()}',
          pgn=PGN(moves=pgn), status='done'
        )
        ses.add(game)
        ses.commit()
      
  def get_images(self, **gameId: Unpack[cp.GameId]) -> list[str] | None:
    with Session(self.engine) as ses:
      game = ses.exec(queries.select_game(**gameId)).first()
      if game and game.upload:
        return [img.descaled_url for img in game.upload.imgs]
      
  async def _upload(self, image: UploadFile, url: str):
    return await self.images.insert(url, await image.read())
  
  async def _descale_imgs(self, imgs: Sequence[tuple[str, bytes]], target_height: int):
    """Descales and uploads the images. Expects `(url, img)` tuples"""
    descaled = [descale2jpg(img, target_height) for _, img in imgs]
    uploads = [self.images.insert(url, img) for (url, _), img in zip(imgs, descaled)]
    return await asyncio.gather(*uploads)

  async def post_game(
    self, images: Sequence[UploadFile],
    descaled_height: int = 768, **gameId: Unpack[cp.GameId]
  ) -> tuple[Either, Coroutine]:
    """ - Runs tasks that must happen before responding to the client.
    - Returns the result + a coroutine that must run after responding (independent of the result)
    """
    img_urls = []
    descaled_urls = []
    delete_urls = []
    imgs = []

    try:
      with Session(self.engine) as ses:
        game = ses.exec(queries.select_game(**gameId)).first()
        if game is None:
          return Left(InexistentItem(detail=f'Game {gameId} not found in DB')), P.of(None).run()
        
        uuid = f'{cp.stringifyId(**gameId)}_{uuid4()}'
        img_urls = [f'{uuid}/{i}.' + (safe_extension(img) or 'jpg') for i, img in enumerate(images)]
        descaled_urls = [f'{uuid}/{i}-{descaled_height}.jpg' for i in range(len(images))]

        imgs = await asyncio.gather(*[img.read() for img in images])
        original_uploads = [self.images.insert(url, img) for url, img in zip(img_urls, imgs)]
        res = E.sequence(await asyncio.gather(*original_uploads))

        if (upl := game.upload):
          delete_urls = [img.url for img in upl.imgs] + [img.descaled_url for img in upl.imgs]
          [ses.delete(img) for img in upl.imgs]

        upload = Upload(
          id=uuid, status='uploaded',
          imgs=[Image(url=url, descaled_url=url) for url in img_urls]
        )
        game.upload = upload
        ses.add(upload)
        ses.add(game)
        ses.commit()

    except Exception as e:
      res = Left(e)

    if res.tag == 'right':
      async def post_tasks():
        deletions = [self.images.delete(url) for url in delete_urls]
        descaled_uploads = self._descale_imgs(list(zip(descaled_urls, imgs)), descaled_height)
        await asyncio.gather(*deletions, descaled_uploads)

        with Session(self.engine) as ses:
          if (game := ses.exec(queries.select_game(**gameId)).first()) and game.upload:
            for img, descaled_url in zip(game.upload.imgs, descaled_urls):
              img.descaled_url = descaled_url
              ses.add(img)
            ses.commit()

    else:
      async def post_tasks():
        await asyncio.gather(*[self.images.delete(url) for url in img_urls]) # not to leak storage

    return res, post_tasks()



from datetime import datetime
from uuid import uuid4
from sqlmodel import Session
import chess_pairings as cp 
import chess_scraping as cs
from moveread.tnmt import Game, Upload, queries
from dslog import Logger

def update_round_pairings(session: Session, *, tournId: str, group: str, round: str, pairings: cp.RoundPairings):
  """Insert/update pairings of a round"""
  round_games = session.exec(queries.select_round_games(**cp.roundId(tournId, group, round))).all()
  games_idx = cp.GamesMapping[Game].from_pairs([(g.gameId(), g) for g in round_games])
  added: list[cp.GameId] = []
  updated: list[cp.GameId] = []

  for idx, pair in enumerate(pairings):
    if pair.tag == 'unpaired':
      continue
    gid = cp.gameId(tournId, group, round, pair.board)
    if gid in games_idx:
      game = games_idx[gid]
      if game.white != pair.white or game.black != pair.black or game.result != pair.result:
        game.white = pair.white; game.black = pair.black; game.result = pair.result # type: ignore
        updated.append(gid)
    else:
      game = Game(
        tournId=tournId, group=group, round=round, board=pair.board, index=idx,
        white=pair.white, black=pair.black, result=pair.result # type: ignore
      )
      if pair.tag == 'team':
        game.upload = Upload(id=str(uuid4()), status='header')
      added.append(gid)
    session.add(game)
  
  session.commit()
  return added, updated
  
def update_group_pairings(session: Session, *, tournId: str, group: str, pairings: cp.GroupPairings):
    """Insert/update pairings of a group"""
    group_games = session.exec(queries.select_group_games(tournId, group)).all()
    games_idx = cp.GamesMapping[Game].from_pairs([(g.gameId(), g) for g in group_games])
    added: list[cp.GameId] = []
    updated: list[cp.GameId] = []

    for round, rnd_pairings in pairings.items():
      for idx, pair in enumerate(rnd_pairings):
        if pair.tag == 'unpaired':
          continue
        gid = cp.gameId(tournId, group, round, pair.board)
        if gid in games_idx:
          game = games_idx[gid]
          if game.white != pair.white or game.black != pair.black or game.result != pair.result:
            game.white = pair.white; game.black = pair.black; game.result = pair.result # type: ignore
            updated.append(gid)
        else:
          game = Game(
            tournId=tournId, group=group, round=round, board=pair.board, index=idx,
            white=pair.white, black=pair.black, result=pair.result # type: ignore
          )
          if pair.tag == 'team':
            game.upload = Upload(id=str(uuid4()), status='header')
          added.append(gid)
        session.add(game)
    session.commit()
    return added, updated


async def scrape_all_pairings(
  session: Session, *, now: datetime | None = None, logger: Logger = Logger.click(),
  scrape_group = cs.scrape_group
):
  """Updates pairings for all ongoing tournaments"""
  pairing_sources = queries.current_pairings(session, now)
  current_tnmts = [tnmt.tournId for tnmt, _ in pairing_sources]
  logger(f'Updating current tournaments: {", ".join(current_tnmts)}')
  for tnmt, src in pairing_sources:
    e = await scrape_group(src.pairings)
    if e.tag == 'left':
      logger(f'Error fetching pairings for "{tnmt.tournId}/{src.group}"', e.value, level='ERROR')
      continue
    added, updated = update_group_pairings(session, tournId=tnmt.tournId, group=src.group, pairings=e.value)
    logger(f'Updated pairings for "{tnmt.tournId}/{src.group}", added {len(added)} and updated {len(updated)} games', level='DEBUG')


async def scrape_current_pairings(
  session: Session, *, now: datetime | None = None, logger: Logger = Logger.click(),
  scrape_round = cs.scrape_round
):
  pairing_sources = queries.current_pairings(session, now)
  current_tnmts = [tnmt.tournId for tnmt, _ in pairing_sources]
  logger(f'Updating current tournaments: {", ".join(current_tnmts)}')
  for tnmt, src in pairing_sources:
    e = queries.current_round(session, tnmt.tournId, src.group, now=now)
    if e.tag == 'left':
      logger(f'Error finding current round for "{tnmt.tournId}/{src.group}":', e.value, level='ERROR')
      continue
    if (rnd := e.value) is None:
      logger(f'No round to poll for "{tnmt.tournId}/{src.group}"')
      continue

    logger(f'Updating current round for {tnmt.tournId} {src.group}: {rnd}')
    e = await scrape_round(src.pairings, rnd)
    if e.tag == 'left':
      logger(f'Error scraping pairings for "{tnmt.tournId}/{src.group}":', e.value, level='WARNING')
      continue
    
    added, updated = update_round_pairings(session, tournId=tnmt.tournId, group=src.group, round=rnd, pairings=e.value)
    logger(f'Updated pairings for "{tnmt.tournId}/{src.group}", added {len(added)} and updated {len(updated)} games', level='DEBUG')


from ._types import Game, GameId, Image, Tournament, PGN, SheetModel, \
  Group, Pairings, Round, FrontendGame, FrontendPGN, Token, Upload
from . import server, queries, jobs, admin

__all__ = [
  'Game', 'GameId', 'Image', 'Tournament', 'PGN', 'SheetModel', 'Upload',
  'Group', 'Pairings', 'Round', 'FrontendGame', 'FrontendPGN', 'Token',
  'server', 'queries', 'jobs', 'admin',
]


from typing import Mapping
from datetime import datetime
from sqlmodel import Session, select
from dslog import Logger
import chess_scraping as cs
from moveread.tnmt import Tournament, Round, SheetModel, queries, Token, Pairings

def upsert_tnmt(session: Session, tnmt: Tournament, *, logger: Logger = Logger.click()):
  if (curr := session.exec(queries.select_tnmt(tnmt.tournId)).first()):
    logger('Deleting existing entry:', curr, level='DEBUG')
    session.delete(curr)
  logger('Inserting new entry:', tnmt)
  session.add(tnmt)

def upsert_round(
  session: Session, tournId: str, group: str, round: str, *,
  logger: Logger = Logger.click(), start: datetime
):
  if not (rnd := session.exec(queries.select_round(tournId, group, round)).first()):
    rnd = Round(tournId=tournId, group=group, name=round, start_dtime=start)
    logger(f'Inserting new round: "{tournId}/{group}/{round}:"', start.strftime('%d/%m/%Y %H:%M:%S'))
    session.add(rnd)
  elif rnd.start_dtime == start:
    logger('Round start time already set to', start)
  else:
    logger('Updating round start time:', rnd.start_dtime, '->', start)
    rnd.start_dtime = start
    session.add(rnd)

def upsert_model(session: Session, tournId: str, model: str, *, logger: Logger = Logger.click()):
  if (curr := session.exec(queries.select_model(tournId)).first()):
    if curr.model == model:
      logger(f'Model "{model}" already set for "{tournId}"')
    else:
      logger(f'Updating model for "{tournId}": "{curr.model}" -> "{model}"')
      curr.model = model
      session.add(curr)
  else:
    logger(f'Inserting new model for "{tournId}": "{model}"')
    session.add(SheetModel(tournId=tournId, model=model))

def upsert_pairings(session: Session, tournId: str, group: str, pairings: cs.Source, *, logger: Logger):
  if (obj := session.get(Pairings, (tournId, group))):
    if obj.pairings.tag == pairings.tag and obj.pairings.id == pairings.id:
      logger(f'Pairings for "{group}" are already up to date')
    else:
      logger(f'Current: {obj.pairings.tag} ({obj.pairings.id}), New: {pairings.tag} ({pairings.id})')
      obj.pairings = pairings
      session.add(obj)
      logger(f'Updated pairings entry for "{group}"')
  else:
    session.add(Pairings(tournId=tournId, group=group, pairings=pairings))
    logger(f'Created pairings entry for "{group}"')


def list_tokens(session: Session, tournId: str):
  stmt = select(Token.token).where(Token.tournId == tournId)
  return session.exec(stmt).all()

def add_token(session: Session, tournId: str, token: str, *, logger: Logger = Logger.click()):
  if session.exec(select(Token).where(Token.tournId == tournId, Token.token == token)).first():
    logger(f'Token "{token}" already exists for "{tournId}"')
  else:
    logger(f'Adding token "{token}" for "{tournId}"')
    session.add(Token(tournId=tournId, token=token))

def revoke_token(session: Session, tournId: str, token: str, *, logger: Logger = Logger.click()):
  stmt = select(Token).where(Token.tournId == tournId, Token.token == token)
  if (tok := session.exec(stmt).first()):
    logger(f'Revoking token "{token}" for "{tournId}"')
    session.delete(tok)
  else:
    logger(f'Token "{token}" not found for "{tournId}"')

class API:
  def __init__(self, session: Session, logger: Logger = Logger.click()):
    self.session = session
    self.logger = logger

  def list_tnmts(self):
    return self.session.exec(select(Tournament.tournId)).all()

  def upsert_tnmt(self, tnmt: Tournament):
    upsert_tnmt(self.session, tnmt, logger=self.logger)

  def upsert_model(self, tournId: str, model: str):
    upsert_model(self.session, tournId, model, logger=self.logger)

  def upsert_round(self, tournId: str, group: str, round: str, start: datetime):
    upsert_round(self.session, tournId, group, round, start=start, logger=self.logger)

  def upsert_pairings(self, tournId: str, group: str, pairings: cs.Source):
    upsert_pairings(self.session, tournId, group, pairings, logger=self.logger)

  def list_tokens(self, tournId: str):
    return list_tokens(self.session, tournId)
  
  def add_token(self, tournId: str, token: str):
    add_token(self.session, tournId, token, logger=self.logger)

  def revoke_token(self, tournId: str, token: str):
    revoke_token(self.session, tournId, token, logger=self.logger)

  def upsert_details(self, tnmt: Tournament, rounds: Mapping[str, Mapping[str, datetime]]):
    self.upsert_tnmt(tnmt)
    for group, rnds in rounds.items():
      for rnd, start in rnds.items():
        self.upsert_round(tnmt.tournId, group, rnd, start)



from typing import Iterable
from chess_utils import PGNHeaders, export_pgn
from .._types import Game, Tournament, Round

def export(game: Game, tnmt: Tournament | None = None, round: Round | None = None) -> str | None:
  site = tnmt and tnmt.site
  event = tnmt and tnmt.name
  dtime = round and round.start_dtime
  headers = PGNHeaders(
    Event=event, Site=site, White=game.white, Black=game.black,
    Round=f'{game.round}.{game.board}', Result=game.result,
    Date=dtime and dtime.strftime('%Y.%m.%d')
  )
  upl = game.upload
  if upl and upl.pgn:
    comment = '[...]' if upl.pgn.early else None
    return export_pgn(upl.pgn.moves, headers, comment)
  else:
    return export_pgn([], headers)

def export_all(games: Iterable[Game], tnmt: Tournament | None = None, round: Round | None = None) -> Iterable[str]:
  for game in games:
    if (pgn := export(game, tnmt, round)) is not None:
      yield pgn + '\n\n'


from typing import Any
from sqlmodel import select
from moveread.tnmt import Game, Upload, Tournament, Round, SheetModel

def group_games(tournId: str, group: str):
  return Game.tournId == tournId, Game.group == group

def group_rounds(tournId: str, group: str):
  return Round.tournId == tournId, Round.group == group

def round_games(tournId: str, group: str, round: str):
  return Game.tournId == tournId, Game.group == group, Game.round == round

def exact_game(tournId: str, group: str, round: str, board: str):
  return Game.tournId == tournId, Game.group == group, Game.round == round, Game.board == board

def exact_round(tournId: str, group: str, round: str):
  return Round.tournId == tournId, Round.group == group, Round.name == round

def select_game(tournId: str, group: str, round: str, board: str):
  return select(Game).where(*exact_game(tournId, group, round, board))

def select_group_games(tournId: str, group: str):
  return select(Game).where(*group_games(tournId, group))

def select_round_games(tournId: str, group: str, round: str):
  return select(Game).where(*round_games(tournId, group, round))

def select_tnmt(tournId: str):
  return select(Tournament).where(Tournament.tournId == tournId)

def select_tnmt_rounds(tournId: str):
  return select(Round).where(Round.tournId == tournId)

def select_round(tournId: str, group: str, round: str):
  return select(Round).where(*exact_round(tournId, group, round))

def select_rounds(tournId: str, group: str):
  """All tournament's rounds, sorted by start dtime"""
  return select(Round).where(Round.tournId == tournId, Round.group == group) \
    .order_by(Round.start_dtime) # type: ignore

def select_model(tournId: str):
  return select(SheetModel).where(SheetModel.tournId == tournId)

def game_uploads():
  eq: Any = Game.uploadId == Upload.id # join has bad types idk
  return select(Game, Upload).join(Upload, eq).where(Upload.status == 'uploaded')


from typing import Sequence, ClassVar, Literal
from dataclasses import dataclass
from datetime import date, datetime
from sqlmodel import Field, SQLModel, Relationship
from sqlalchemy import String
from sqltypes import SpaceDelimitedList, ValidatedJSON
from chess_pairings import GameId, gameId, Result
from chess_scraping import Source

class SheetModel(SQLModel, table=True):
  tournId: str = Field(primary_key=True, foreign_key='tournament.tournId')
  model: str

class Tournament(SQLModel, table=True):
  tournId: str = Field(primary_key=True)
  name: str
  site: str | None = None
  start_date: date
  end_date: date
  groups: Sequence[str] = Field(sa_type=SpaceDelimitedList)

  def __repr__(self):
    import json
    dct = dict(self)
    dct['start_date'] = self.start_date.strftime('%d/%m/%Y')
    dct['end_date'] = self.end_date.strftime('%d/%m/%Y')
    return 'Tournament(' + json.dumps(dct, indent=2) + ')'
  __str__ = __repr__
  

class Pairings(SQLModel, table=True):
  tournId: str = Field(foreign_key='tournament.tournId', primary_key=True)
  group: str = Field(primary_key=True)
  pairings: Source = Field(sa_type=ValidatedJSON(Source))

@dataclass
class Group:
  tournId: str
  name: str
  rounds: Sequence[str]

class Round(SQLModel, table=True):
  tournId: str = Field(primary_key=True, foreign_key='tournament.tournId')
  group: str = Field(primary_key=True)
  name: str = Field(primary_key=True)
  start_dtime: datetime
  
class Image(SQLModel, table=True):
  id: int | None = Field(default=None, primary_key=True)
  url: str
  descaled_url: str
  uploadId: str = Field(default=None, foreign_key='upload.id')

class FrontendPGN(SQLModel):
  moves: Sequence[str] = Field(sa_type=SpaceDelimitedList)
  early: bool | None = None

class PGN(FrontendPGN, table=True):
  uploadId: str = Field(default=None, primary_key=True, foreign_key='upload.id')

@dataclass
class Paired:
  white: str
  black: str
  result: Result | None = None


@dataclass
class FrontendGame:
  board: str
  pairing: Paired
  status: 'GameStatus | None' = None

  @classmethod
  def of(cls, game: 'Game'):
    return cls(game.board, Paired(white=game.white, black=game.black, result=game.result), game.status)

class Upload(SQLModel, table=True):
  id: str = Field(primary_key=True)
  imgs: list[Image] = Relationship()
  pgn: PGN | None = Relationship()
  games: list['Game'] = Relationship(back_populates='upload')
  status: 'GameStatus | None' = Field(default=None, sa_type=String)

GameStatus = Literal['uploaded', 'doing', 'done', 'header']
"""header: the game is not a real game but a header for team tournaments"""

class Game(SQLModel, table=True):
  Status: ClassVar = GameStatus
  id: int | None = Field(default=None, primary_key=True)
  tournId: str = Field(foreign_key='tournament.tournId')
  uploadId: str | None = Field(default=None, foreign_key='upload.id')
  group: str
  round: str
  board: str
  index: int
  """Boards may have out of whack names (e.g. in team tournaments "1.3"). This is the order you'd see in chess-results"""
  white: str
  black: str
  result: Result | None = Field(sa_type=String)
  upload: Upload | None = Relationship(back_populates='games')

  @property
  def status(self) -> GameStatus | None:
    return self.upload and self.upload.status

  def gameId(self) -> GameId:
    return gameId(self.tournId, self.group, self.round, self.board)
  
class Token(SQLModel, table=True):
  id: int | None = Field(default=None, primary_key=True)
  token: str
  tournId: str = Field(foreign_key='tournament.tournId')
  
__all__ = [
  'Game', 'GameId', 'Image', 'Tournament', 'PGN', 'Paired', 'Token'
]


from typing import Literal
import asyncio
from datetime import datetime, timedelta
from fastapi import FastAPI, Response, UploadFile
from fastapi.responses import StreamingResponse
from kv import LocatableKV
from dslog import Logger
from dslog.uvicorn import setup_loggers_lifespan, DEFAULT_FORMATTER, ACCESS_FORMATTER
import chess_pairings as cp
from moveread.tnmt import Tournament, Group, FrontendGame, FrontendPGN
from .sdk import SDK
from .util import notify, Bearer

def fastapi(
  sdk: SDK, images: LocatableKV[bytes], *,
  logger: Logger = Logger.click().prefix('[TNMT API]'),
):

  app = FastAPI(
    generate_unique_id_function=lambda route: route.name,
    lifespan=setup_loggers_lifespan(
      access=logger.format(ACCESS_FORMATTER),
      uvicorn=logger.format(DEFAULT_FORMATTER),
    )
  )

  @app.get('/authorize/{tournId}', responses={ 401: {}, 200: {} })
  def authorize(tournId: str, token: str = Bearer):
    authed = sdk.authorize(token, tournId)
    return Response(status_code=200 if authed else 401)

  @app.get('/')
  def get_tournaments(skip: int = 0, take: int = 10) -> list[Tournament]:
    return list(sdk.get_tnmts(skip, take))

  @app.get('/{tournId}', responses={ 404: {}, 200: { 'model': Tournament }})
  def get_tournament(tournId: str, resp: Response) -> Tournament | None:
    tnmt = sdk.get_tnmt(tournId)
    if tnmt is None:
      resp.status_code = 404
    return tnmt

  @app.get('/{tournId}/{group}.pgn', response_model=str)
  def get_group_pgn(tournId: str, group: str, skip: int = 0, take: int | None = None, no_cache: bool = False):
    pgn = sdk.get_group_pgn(**cp.groupId(tournId, group), skip=skip, take=take, no_cache=no_cache)
    return StreamingResponse(content=pgn, media_type='application/x-chess-pgn', headers={
      'Content-Disposition': f'attachment; filename={tournId}_{group}.pgn'
    })

  @app.get('/{tournId}/{group}/', responses={ 404: {}, 200: { 'model': Group }})
  def get_group(tournId: str, group: str, r: Response) -> Group | None:
    grp = sdk.get_group(tournId, group)
    if grp is None:
      r.status_code = 404
    return grp

  @app.get('/{tournId}/{group}/{round}.pgn', response_model=str)
  def get_round_pgn(tournId: str, group: str, round: str, skip: int = 0, take: int | None = None, no_cache: bool = False):
    pgn = sdk.get_round_pgn(**cp.roundId(tournId, group, round), skip=skip, take=take, no_cache=no_cache)
    return StreamingResponse(content=pgn, media_type='application/x-chess-pgn', headers={
      'Content-Disposition': f'attachment; filename={tournId}_{group}_{round}.pgn'
    })
  
  @app.get('/{tournId}/{group}/{round}')
  def get_round(tournId: str, group: str, round: str) -> list[FrontendGame]:
    return list(sdk.get_round(**cp.roundId(tournId, group, round)))

  @app.get('/{tournId}/{group}/{round}/{board}/pgn')
  def get_pgn(tournId: str, group: str, round: str, board: str, resp: Response) -> FrontendPGN | None:
    pgn = sdk.get_pgn(**cp.gameId(tournId, group, round, board))
    # if pgn is None:
      # resp.status_code = 404
    return pgn

  @app.get('/{tournId}/{group}/{round}/{board}/images', responses={ 404: {}, 200: { 'model': list[str] }})
  def get_images(tournId: str, group: str, round: str, board: str, resp: Response) -> list[str] | None:
    urls = sdk.get_images(**cp.gameId(tournId, group, round, board))
    if urls is None or len(urls) == 0:
      resp.status_code = 404
      return None
    
    expiry = datetime.now() + timedelta(hours=1)
    return [images.url(url, expiry=expiry) for url in urls]

  # @app.post('/{tournId}/{group}/{round}/{board}/move', responses={
  #   401: dict(model=Literal['UNAUTHORIZED']),
  #   200: dict(model=Literal['OK']),
  #   500: dict(model=Literal['ERROR'])
  # })
  # def move_game(
  #   tournId: str, group: str, round: str, board: str,
  #   new_board: str, resp: Response, token: str = Bearer,
  # ) -> Literal['OK', 'UNAUTHORIZED', 'ERROR']:
  #   if not sdk.authorize(token, tournId):
  #     resp.status_code = 401
  #     return 'UNAUTHORIZED'
    
  #   from_id = cp.gameId(tournId, group, round, board)
  #   to_id = cp.gameId(tournId, group, round, new_board)
  #   try:
  #     post = sdk.move_game(from_id, to_id)
  #     asyncio.create_task(post)
  #     return 'OK'
  #   except Exception as e:
  #     logger(f'Error moving "{cp.stringifyId(**from_id)}" -> "{cp.stringifyId(**to_id)}":', e, level='ERROR')
  #     resp.status_code = 500
  #     return 'ERROR'


  @app.post('/{tournId}/{group}/{round}/{board}/pgn', responses={
    401: dict(model=Literal['UNAUTHORIZED']),
    200: dict(model=Literal['OK']),
    500: dict(model=Literal['ERROR'])
  })
  async def post_pgn(
    tournId: str, group: str, round: str, board: str,
    pgn: list[str], r: Response, token: str = Bearer
  ) -> Literal['OK', 'UNAUTHORIZED', 'ERROR']:
    
    if not sdk.authorize(token, tournId):
      r.status_code = 401
      return 'UNAUTHORIZED'
    
    gid = cp.gameId(tournId, group, round, board)
    try:
      sdk.post_pgn(pgn, **gid)
      return 'OK'
    except Exception as e:
      logger(f'Error posting PGN "{cp.stringifyId(**gid)}":', e, level='ERROR')
      r.status_code = 500
      return 'ERROR'

  @app.post('/{tournId}/{group}/{round}/{board}', responses={
    401: dict(model=Literal['UNAUTHORIZED']),
    200: dict(model=Literal['OK']),
    500: dict(model=Literal['ERROR'])
  })
  async def post_game(
    tournId: str, group: str, round: str, board: str,
    images: list[UploadFile], r: Response, token: str = Bearer
  ) -> Literal['OK', 'UNAUTHORIZED', 'ERROR']:
  
    if not sdk.authorize(token, tournId):
      r.status_code = 401
      return 'UNAUTHORIZED'
    
    gid = cp.gameId(tournId, group, round, board)
    res, coro = await sdk.post_game(images, **gid)

    async def post_task():
      r = await coro
      logger('Post task finished, returning:', r)
    asyncio.create_task(post_task())
    
    if res.tag == 'left':
      logger(f'Error posting game "{cp.stringifyId(**gid)}":', res.value, level='ERROR')
      r.status_code = 500
      return 'ERROR'
    
    asyncio.create_task(notify(gid, logger))
    return 'OK' 

  # @app.delete('/{tournId}/{group}/{round}/{board}', responses={
  #   401: dict(model=Literal['UNAUTHORIZED']),
  #   500: dict(model=Literal['ERROR']),
  #   200: dict(model=Literal['OK']),
  # })
  # def delete_game(
  #   tournId: str, group: str, round: str, board: str,
  #   resp: Response, token: str = Bearer
  # ):
  #   if not sdk.authorize(token, tournId):
  #     resp.status_code = 401
  #     return 'UNAUTHORIZED'
    
  #   gid = cp.gameId(tournId, group, round, board)
  #   try:
  #     post = sdk.delete_game(**gid)
  #     asyncio.create_task(post)
  #     return 'OK'
  #   except Exception as e:
  #     logger(f'Error deleting game "{cp.stringifyId(**gid)}":', e, level='ERROR')
  #     resp.status_code = 500
  #     return 'ERROR'

  return app


import os
from argparse import ArgumentParser
from openapi_ts import generate_client
from moveread.tnmt.server import fastapi

def main():

  parser = ArgumentParser()
  parser.add_argument('-o', '--output', required=True)
  args = parser.parse_args()

  app = fastapi({}, {}) # type: ignore
  spec = app.openapi()
  generate_client(spec, args.output, args={
    '--client': '@hey-api/client-fetch',
    '--services': '{ asClass: false }',
    '--schemas': 'false'
  })

if __name__ == '__main__':
  main()


import typer
from sqlmodel import create_engine, Session
from moveread.tnmt import jobs

main = typer.Typer()

pairings = typer.Typer()
main.add_typer(pairings, name='pairings')

@pairings.command()
async def current(
  sql: str = typer.Option(help='SQL connection string', envvar='SQL_CONN_STR')
):
  engine = create_engine(sql)
  with Session(engine) as session:
    await jobs.scrape_current_pairings(session)

@pairings.command()
async def all(
  sql: str = typer.Option(help='SQL connection string', envvar='SQL_CONN_STR')
):
  engine = create_engine(sql)
  with Session(engine) as session:
    await jobs.scrape_all_pairings(session)


from .basic import group_games, round_games, exact_game, exact_round, \
  select_game, select_group_games, select_tnmt, select_round, \
  select_rounds, select_tnmt_rounds, select_model, \
  game_uploads, group_rounds, select_round_games
from .pairings import round_paired, current_pairings, current_round, round_finished, \
  select_paired_rounds
from .api import API, Select

__all__ = [
  'group_games', 'round_games', 'exact_game', 'exact_round',
  'select_game', 'select_group_games', 'select_tnmt', 'select_round',
  'select_rounds', 'select_tnmt_rounds', 'select_model',
  'game_uploads', 'group_rounds', 'select_round_games',
  'round_paired', 'current_pairings', 'current_round', 'round_finished',
  'select_paired_rounds', 'API', 'Select',
]


import os

DEFAULT_ORIGINS = [
  'http://localhost:5173',
  'http://localhost:4713',
  'https://localhost',
  'https://moveread.com',
  'https://dfy.moveread.com',
  'https://tnmt.moveread.com',
]

def get_env(variable: str, default=None, required=True):
  value = os.getenv(variable, default)
  if value is None and required:
      raise ValueError(f"Environment variable {variable} is required.")
  return value

# Check environment variables first
if get_env('LOAD_DOTENV', required=False):
  from dotenv import load_dotenv
  load_dotenv()
   
blob_conn_str = get_env('BLOB_CONN_STR')
sql_conn_str = get_env('SQL_CONN_STR')
images_path = get_env('IMAGES_PATH', default=None, required=False)
cache_ttl = float(get_env('CACHE_TTL', default=60*5)) # type: ignore
cache_max_entries = int(get_env('CACHE_MAX_ENTRIES', default=1000)) # type: ignore
port = int(get_env('PORT', default=80)) # type: ignore
host = get_env('HOST', default='0.0.0.0')
cors_origins = get_env('CORS_ORIGINS', default=",".join(DEFAULT_ORIGINS)).split(",") # type: ignore

# Import other modules after checking environment variables
from dslog import Logger, formatters
from kv import KV, LocatableKV
from sqlmodel import create_engine
from fastapi.middleware.cors import CORSMiddleware
from moveread.tnmt import server

logger = Logger.stderr().format(formatters.click)
logger('Starting API...')
logger('- Cache TTL: ', cache_ttl)
logger('- Cache Max Entries: ', cache_max_entries)
logger('- CORS allowed origins: ', cors_origins)

blobs = KV.of(blob_conn_str) # type: ignore
assert isinstance(blobs, LocatableKV)

engine = create_engine(sql_conn_str) # type: ignore

sdk = server.SDK(blobs, engine, cache_ttl=cache_ttl, cache_max_entries=cache_max_entries)
app = server.fastapi(sdk, images=blobs, logger=logger)
app.add_middleware(CORSMiddleware, allow_origins=cors_origins, allow_credentials=True, allow_methods=["*"], allow_headers=["*"])
